import jax
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.interpreters.xla as xla
import jax.numpy as jnp


# unvmap_all

_unvmap_all_p = jax.core.Primitive("unvmap_all")


def unvmap_all(x):
    """As `jnp.all`, but ignores batch dimensions."""
    return _unvmap_all_p.bind(x)


def _unvmap_all_impl(x):
    return jnp.all(x)


def _unvmap_all_abstract_eval(x):
    return jax.ShapedArray(shape=(), dtype=jax.numpy.bool_.dtype)


def _unvmap_all_batch(x, batch_axes):
    (x,) = x
    return unvmap_all(x), batching.not_mapped


_unvmap_all_p.def_impl(_unvmap_all_impl)
_unvmap_all_p.def_abstract_eval(_unvmap_all_abstract_eval)
batching.primitive_batchers[_unvmap_all_p] = _unvmap_all_batch
if hasattr(xla, "lower_fun"):
    xla.register_translation(
        _unvmap_all_p,
        xla.lower_fun(_unvmap_all_impl, multiple_results=False, new_style=True),
    )
mlir.register_lowering(
    _unvmap_all_p,
    mlir.lower_fun(_unvmap_all_impl, multiple_results=False),
)

# unvmap_any

_unvmap_any_p = jax.core.Primitive("unvmap_any")


def unvmap_any(x):
    """As `jnp.any`, but ignores batch dimensions."""
    return _unvmap_any_p.bind(x)


def _unvmap_any_impl(x):
    return jnp.any(x)


def _unvmap_any_abstract_eval(x):
    return jax.ShapedArray(shape=(), dtype=jax.numpy.bool_.dtype)


def _unvmap_any_batch(x, batch_axes):
    (x,) = x
    return unvmap_any(x), batching.not_mapped


_unvmap_any_p.def_impl(_unvmap_any_impl)
_unvmap_any_p.def_abstract_eval(_unvmap_any_abstract_eval)
batching.primitive_batchers[_unvmap_any_p] = _unvmap_any_batch
if hasattr(xla, "lower_fun"):
    xla.register_translation(
        _unvmap_any_p,
        xla.lower_fun(_unvmap_any_impl, multiple_results=False, new_style=True),
    )
mlir.register_lowering(
    _unvmap_any_p,
    mlir.lower_fun(_unvmap_any_impl, multiple_results=False),
)

# unvmap_max

_unvmap_max_p = jax.core.Primitive("unvmap_max")


def unvmap_max(x):
    """As `jnp.max`, but ignores batch dimensions."""
    return _unvmap_max_p.bind(x)


def _unvmap_max_impl(x):
    return jnp.max(x)


def _unvmap_max_abstract_eval(x):
    return jax.ShapedArray(shape=(), dtype=x.dtype)


def _unvmap_max_batch(x, batch_axes):
    (x,) = x
    return unvmap_max(x), batching.not_mapped


_unvmap_max_p.def_impl(_unvmap_max_impl)
_unvmap_max_p.def_abstract_eval(_unvmap_max_abstract_eval)
batching.primitive_batchers[_unvmap_max_p] = _unvmap_max_batch
if hasattr(xla, "lower_fun"):
    xla.register_translation(
        _unvmap_max_p,
        xla.lower_fun(_unvmap_max_impl, multiple_results=False, new_style=True),
    )
mlir.register_lowering(
    _unvmap_max_p,
    mlir.lower_fun(_unvmap_max_impl, multiple_results=False),
)



import typing
from typing import Any, Callable


# Inherits from type so that _WithRepr instances are types and can be used as
# e.g. Sequence[_WithRepr(...)]
class _WithRepr(type):
    def __new__(self, string):
        out = super().__new__(self, string, (), {})
        # prevent the custom typing repr from doing the wrong thing
        out.__module__ = "builtins"
        return out

    def __init__(self, string):
        self.string = string

    def __repr__(self):
        return self.string


def doc_repr(obj: Any, string: str):
    if getattr(typing, "GENERATING_DOCUMENTATION", False):
        return _WithRepr(string)
    else:
        return obj


def doc_strip_annotations(fn: Callable) -> Callable:
    if getattr(typing, "GENERATING_DOCUMENTATION", False):
        fn.__annotations__ = None
    return fn



from typing import Sequence, Union

import jax
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.lax as lax
import jax.tree_util as jtu
from jaxtyping import Array, Bool, Int, PyTree

from ..filters import combine, is_array_like, partition
from .unvmap import unvmap_any, unvmap_max


def _error_impl(pred, index, *x, msgs):
    def raises(_index):
        raise RuntimeError(msgs[_index.item()])

    struct = jax.eval_shape(lambda: x)
    return lax.cond(pred, lambda: jax.pure_callback(raises, struct, index), lambda: x)


def _error_abstract(pred, index, *x, msgs):
    return x


def _error_jvp(primals, tangents, *, msgs):
    _, _, *tx = tangents
    return _branched_error_p.bind(*primals, msgs=msgs), tx


def _error_transpose(cts, pred, index, *x, msgs):
    return (None, None) + cts


def _error_batching(inputs, batch_axes, *, msgs):
    pred_bdim, index_bdim, *xs_bdim = batch_axes
    assert pred_bdim is batching.not_mapped
    assert index_bdim is batching.not_mapped
    return _branched_error_p.bind(*inputs, msgs=msgs), xs_bdim


_branched_error_p = jax.core.Primitive("branched_error")
_branched_error_p.multiple_results = True
_branched_error_p.def_impl(_error_impl)
_branched_error_p.def_abstract_eval(_error_abstract)
ad.primitive_jvps[_branched_error_p] = _error_jvp
ad.primitive_transposes[_branched_error_p] = _error_transpose
batching.primitive_batchers[_branched_error_p] = _error_batching
mlir.register_lowering(
    _branched_error_p, mlir.lower_fun(_error_impl, multiple_results=True)
)


def error_if(
    x: PyTree,
    pred: Union[bool, Bool[Array, "..."]],
    msg: str,
) -> PyTree:
    """Throws an error based on runtime values.
    Works even under JIT.

    Note that this probably won't work on TPU (a dummy value
    is propagated through the computation instead of an error
    being thrown).

    The first argument `x` is returned unchanged, and used to determine
    where the error check happens in the overall computation. If the
    return value is ever unused then the error check may be DCE'd.

    !!! Example

        ```python
        @jax.jit
        def f(x):
            x = error_if(x, x < 0, "x must be >= 0")
            # ...use x in your computation...
            return x

        f(jax.numpy.array(-1))
        ```
    """
    return branched_error_if(x, pred, 0, [msg])


def branched_error_if(
    x: PyTree,
    pred: Union[bool, Bool[Array, "..."]],
    index: Union[int, Int[Array, "..."]],
    msgs: Sequence[str],
) -> PyTree:
    """As [`equinox.internal.error_if`][], but will raise one of
    several messages depending on the value of `index`.
    """

    with jax.ensure_compile_time_eval():
        pred = unvmap_any(pred)
        if not isinstance(pred, jax.core.Tracer) and pred.item() is False:
            return x

    index = unvmap_max(index)
    dynamic_x, static_x = partition(x, is_array_like)
    flat, treedef = jtu.tree_flatten(dynamic_x)
    if len(flat) == 0:
        raise ValueError("No array-likes to thread error on to")
    # Use a primitive to put the lax.cond inside the impl rule.
    # This is needed as lax.cond will unnecessarily promote symbolic
    # zeros to non-symbolic-zeros, and we'd really like to avoid that.
    flat = _branched_error_p.bind(pred, index, *flat, msgs=msgs)
    dynamic_x = jtu.tree_unflatten(treedef, flat)
    return combine(dynamic_x, static_x)



import functools as ft
import types
import typing
import warnings
from typing import Any, Callable, Dict

import jax
import jax.interpreters.ad as ad
import jax.tree_util as jtu
from jaxtyping import Array, PyTree

from .custom_types import BoolAxisSpec, sentinel
from .doc_utils import doc_strip_annotations
from .filters import (
    combine,
    is_array,
    is_inexact_array,
    is_inexact_array_like,
    partition,
)
from .make_jaxpr import filter_make_jaxpr
from .module import Module, module_update_wrapper, Static


class _ValueAndGradWrapper(Module):
    _fun: Callable
    _arg: PyTree[BoolAxisSpec]
    _gradkwargs: Dict[str, Any]

    # Try to avoid clashes with existing argument names.
    # TODO: use "/" once we're on Python 3.8.
    def __call__(__self, __x, *args, **kwargs):
        @ft.partial(jax.value_and_grad, argnums=0, **__self._gradkwargs)
        def fun_value_and_grad(_diff_x, _nondiff_x, *_args, **_kwargs):
            _x = combine(_diff_x, _nondiff_x)
            return __self._fun(_x, *_args, **_kwargs)

        diff_x, nondiff_x = partition(__x, __self._arg)
        return fun_value_and_grad(diff_x, nondiff_x, *args, **kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return jtu.Partial(self, instance)


class _GradWrapper(Module):
    _fun_value_and_grad: _ValueAndGradWrapper
    _has_aux: bool

    def __call__(__self, *args, **kwargs):
        value, grad = __self._fun_value_and_grad(*args, **kwargs)
        if __self._has_aux:
            _, aux = value
            return grad, aux
        else:
            return grad

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return jtu.Partial(self, instance)


@doc_strip_annotations
def filter_value_and_grad(
    fun: Callable = sentinel,
    *,
    arg: PyTree[BoolAxisSpec] = is_inexact_array,
    **gradkwargs,
) -> Callable:
    """As [`equinox.filter_grad`][], except that it is `jax.value_and_grad` that is
    wrapped.
    """

    if fun is sentinel:
        return ft.partial(filter_value_and_grad, arg=arg, **gradkwargs)

    filter_spec = gradkwargs.pop("filter_spec", None)
    if filter_spec is not None:
        warnings.warn("For brevity the `filter_spec` argument has been renamed `arg`")
        arg = filter_spec

    argnums = gradkwargs.pop("argnums", None)
    if argnums is not None:
        raise ValueError(
            "`argnums` should not be passed. If you need to differentiate "
            "multiple objects then collect them into a tuple and pass that "
            "as the first argument."
        )

    return module_update_wrapper(_ValueAndGradWrapper(fun, arg, gradkwargs), fun)


@doc_strip_annotations
def filter_grad(
    fun: Callable = sentinel,
    *,
    arg: PyTree[BoolAxisSpec] = is_inexact_array,
    **gradkwargs,
):
    """As `jax.grad`, but accepts arbitrary PyTrees as inputs. (Not just JAXable types.)

    !!! info

        By default, all inexact (floating-point) JAX arrays are differentiated. Any
        nondifferentiable leaves will have `None` as the gradient.


    **Arguments:**

    - `fun` is a pure function to JIT compile.
    - `arg` is a PyTree whose structure should be a prefix of the structure of
        the **first** argument to `fun`. It behaves as the `filter_spec` argument to
        [`equinox.filter`][]. Truthy values will be differentiated; falsey values will
        not.
    - `**gradkwargs` are any other keyword arguments to `jax.grad`.

    **Returns:**

    A function computing the derivative of `fun` with respect to its first input. Any
    nondifferentiable leaves will have `None` as the gradient. See
    [`equinox.apply_updates`][] for a convenience function that will only attempt to
    apply non-`None` updates.

    !!! tip

        If you need to differentiate multiple objects, then put them together into a
        tuple and pass that through the first argument:
        ```python
        # We want to differentiate `func` with respect to both `x` and `y`.
        def func(x, y):
            ...

        @equinox.filter_grad
        def grad_func(x__y):
            x, y = x__y
            return func(x, y)
        ```
    """

    if fun is sentinel:
        return ft.partial(filter_grad, arg=arg, **gradkwargs)

    has_aux = gradkwargs.get("has_aux", False)

    fun_value_and_grad = filter_value_and_grad(fun, arg=arg, **gradkwargs)
    return module_update_wrapper(_GradWrapper(fun_value_and_grad, has_aux), fun)


def _is_none(x):
    return x is None


def _is_jvp_tracer(x):
    return isinstance(x, ad.JVPTracer)


def filter_jvp(fn, primals, tangents):
    """Like `jax.jvp`, but accepts arbitrary PyTrees. (Not just JAXable types.)

    **Arguments:**

    - `fn`: Function to be differentiated. Its arguments can be Python objects, and
        its return type can be any Python object.
    - `primals`: The primal values at which `fn` should be evaluated. Should be a
        sequence of arguments, and its length should be equal to the number of
        positional parameter of `fn`.
    - `tangents`: The tangent vector for which the Jacobian-vector product should be
        calculated. Should be a PyTree with the same structure as `primals`. The leaves
        of `tangents` must be either floating-point JAX arrays, or Python floats, or
        `None`s. The tangent must be `None` for any primal which is not itself a
        floating-point JAX array or Python float.

    **Returns:**

    A pair `(primals_out, tangents_out)` is returned,
    where `primals_out = fn(*primals)` and `tangents_out` is the Jacobian-vector
    product of `fn` evaluated at `primals` with `tangents`.

    The `tangents_out` has the same structure as `primals_out`, but has `None` for
    any leaves that aren't differentiable.

    !!! Tip

        Unlike `jax.jvp`, this function does not support a `has_aux` argument. It isn't
        needed, as unlike `jax.jvp` the output of this function can be of arbitrary type.
    """
    if jtu.tree_structure(primals, is_leaf=_is_none) != jtu.tree_structure(
        tangents, is_leaf=_is_none
    ):
        raise ValueError("primals and tangents must have the same pytree structure")
    filter_spec = jtu.tree_map(_is_none, tangents, is_leaf=_is_none)
    static_primals, dynamic_primals = partition(primals, filter_spec)
    flat_dynamic_primals, treedef = jtu.tree_flatten(dynamic_primals)
    flat_tangents = jtu.tree_leaves(tangents)  # all non-None tangents are dynamic

    def _fn(*_flat_dynamic):
        _dynamic = jtu.tree_unflatten(treedef, _flat_dynamic)
        _in = combine(_dynamic, static_primals)
        _out = fn(*_in)
        _dynamic_out, _static_out = partition(_out, _is_jvp_tracer)
        return _dynamic_out, Static(_static_out)

    primal_out, tangent_out = jax.jvp(_fn, flat_dynamic_primals, flat_tangents)
    dynamic_primal_out, static_primal_out = primal_out
    primal_out = combine(dynamic_primal_out, static_primal_out.value)
    tangent_out, _ = tangent_out

    return primal_out, tangent_out


def filter_vjp(fun, *primals, has_aux=False):
    """Filtered version of `jax.vjp`.

    **Arguments:**

    - `fun`: The function to be differentiated. Will be called as `fun(*primals)`. Can
        return an arbitrary PyTree.
    - `primals`: The arguments at which `fun` will be evaluated and differentiated.
        Can be arbitrary PyTrees.
    - `has_aux`: Indicates whether `fun` returns a pair, with the first element the
        output to be differentiated, and the latter auxiliary data. Defaults to `False`.

    **Returns:**

    If `has_aux is False` then returns a `(primals_out, vjpfun)` pair, where
    `primals_out = fun(*primals)` and `vjpfun` is a function from a cotangent vector
    with the same shape as `primals_out` to a tuple of cotangent vectors with the same
    shape as `primals`, representing the vector-Jacobian product of `fun` evaluated at
    `primals`.

    If `has_aux is True` then returns a tuple `(primals_out, vjpfun, aux)`, where `aux`
    is the auxiliary data returned from `fun`.

    The cotangent passed to `vjpfun` should have arrays corresponding to all
    floating-point arrays in `primals_out`, and `None` for all other PyTree leaves. The
    cotangents returned from `vjpfun` will likewise have arrays for all `primals` that
    are floating-point arrays, and `None` for all other PyTree leaves.
    """
    diff, nondiff = partition(primals, is_inexact_array)

    def diff_fun(*_diff):
        _primals = combine(_diff, nondiff)
        _out = fun(*_primals)
        if has_aux:
            _out, _aux = _out
        else:
            _aux = None
        _diff_out, _nondiff_out = partition(_out, is_inexact_array)
        return _diff_out, (_nondiff_out, _aux)

    diff_out, vjp_fn, (nondiff_out, aux) = jax.vjp(diff_fun, *diff, has_aux=True)
    out = combine(diff_out, nondiff_out)
    if has_aux:
        return out, vjp_fn, aux
    else:
        return out, vjp_fn


class _ClosureConvert(Module):
    jaxpr: jax.core.Jaxpr
    consts: PyTree[Array]  # Captured in the PyTree structure of _ClosureConvert
    out_dynamic_struct: PyTree[jax.ShapeDtypeStruct]
    out_static: PyTree[Any]

    def __call__(self, *args, **kwargs):
        dynamic = filter((args, kwargs), is_array)
        dynamic_flat = jtu.tree_leaves(dynamic)
        out_dynamic_flat = jax.core.eval_jaxpr(self.jaxpr, self.consts, *dynamic_flat)
        out_dynamic_struct_flat, out_dynamic_treedef = jtu.tree_flatten(
            self.out_dynamic_struct
        )
        assert len(out_dynamic_flat) == len(out_dynamic_struct_flat)
        for o1, o2 in zip(out_dynamic_flat, out_dynamic_struct_flat):
            assert o1.shape == o2.shape
            assert o1.dtype == o2.dtype
        out = jtu.tree_unflatten(out_dynamic_treedef, out_dynamic_flat)
        out = combine(out, self.out_static)
        return out


def filter_closure_convert(fn, *args, **kwargs):
    """As `jax.closure_convert`, but works on functions accepting and returning
    arbtirary PyTree objects. In addition, all JAX arrays are hoisted into constants
    (not just floating point arrays).

    This is useful for explicitly capturing any closed-over JAX tracers
    before crossing an API boundary, such as `jax.grad`, `jax.custom_vjp`, or the
    rule of a custom primitive.

    **Arguments:**

    - `fn`: The function to call. Will be called as `fun(*args, **kwargs)`.
    - `args`, `kwargs`: Example arguments at which to call the function. The function is
        not actually evaluated on these arguments; all JAX arrays are subsituted for
        tracers. Note that Python builtins (`bool`, `int`, `float`, `complex`) are
        not substituted for tracers and are passed through as-is.

    **Returns:**

    A new function, which can be called in the same way, using `*args` and `**kwargs`.
    Will contain all closed-over tracers of `fn` as part of its PyTree structure.

    !!! Example

        ```python
        @jax.grad
        def f(x, y):
            z = x + y
            g = lambda a: z + a  # closes over z
            g2 = filter_closure_convert(g, 1)
            assert [id(b) for b in g2.consts] == [id(z)]
            return z

        f(1., 1.)
        ```
    """
    if fn.__closure__ is None:
        # In this case, it's not possible to have any closed-over tracers.
        return fn
    closed_jaxpr, out_dynamic_struct, out_static = filter_make_jaxpr(fn)(
        *args, **kwargs
    )
    jaxpr = closed_jaxpr.jaxpr
    consts = closed_jaxpr.consts
    return _ClosureConvert(jaxpr, consts, out_dynamic_struct, out_static)


class filter_custom_jvp:
    """Filtered version of `jax.custom_jvp`.

    Works in the same way as `jax.custom_jvp`, except that you do not need to specify
    `nondiff_argnums`. Instead, arguments are automatically split into differentiable
    and nondifferentiable based on whether or not they are a floating-point JAX array.

    The tangents of the nondifferentiable arguments will be passed as `None`.

    The return types must still all be JAX types.

    Example:
    ```python
    @equinox.filter_custom_jvp
    def call(fn, x):
        return fn(x)

    @call.defjvp
    def call_jvp(primals, tangents):
        fn, x = primals
        _, tx = tangents
        primal_out = call(fn, x)
        tangent_out = tx**2
        return primal_out, tangent_out
    ```
    """

    def __init__(self, fn):
        def fn_wrapper(static, dynamic):
            return fn(*combine(dynamic, static))

        self.fn = jax.custom_jvp(fn_wrapper, nondiff_argnums=(0,))

    def defjvp(self, fn_jvp):
        def fn_jvp_wrapper(static, dynamic, tangents):
            (dynamic,) = dynamic
            (tangents,) = tangents
            primals = combine(dynamic, static)
            return fn_jvp(primals, tangents)

        self.fn.defjvp(fn_jvp_wrapper)

    def defjvps(self, *a, **kw):
        raise NotImplementedError("filter_custom_jvp().defjvps is not implemented")

    def __call__(self, *args):
        dynamic, static = partition(args, is_inexact_array_like)
        return self.fn(static, dynamic)


class filter_custom_vjp:
    """As `jax.custom_vjp`, but with a nicer interface.

    Usage is:
    ```python
    @equinox.filter_custom_vjp
    def fn(vjp_arg, *args, **kwargs):
        # vjp_arg is some PyTree of arbitrary Python objects.
        # args, kwargs contain arbitrary Python objects.
        ...
        return obj  # some PyTree of arbitrary Python objects.

    def fn_fwd(vjp_arg, *args, **kwargs):
        ...
        # Should return `obj` as before. `residuals` can be any collection of JAX
        # arrays you want to keep around for the backward pass.
        return obj, residuals

    def fn_bwd(residuals, grad_obj, vjp_arg, *args, **kwargs):
        # grad_obj will have `None` as the gradient for any leaves of `obj` that were
        # not JAX arrays
        ...
        # grad_vjp_arg should have `None` as the gradient for any leaves of `vjp_arg`
        # that were not JAX arrays.
        return grad_vjp_arg

    fn.defvjp(fn_fwd, fn_bwd)
    ```

    The key differences to `jax.custom_vjp` are that:

    - Only the gradient of the first argument, `vjp_arg`, should be computed on the
        backward pass. Everything else will automatically have zero gradient.
    - You do not need to distinguish differentiable from nondifferentiable manually.
        Instead you should return gradients for all inexact JAX arrays in the first
        argument. (And just put `None` on every other leaf of the PyTree.)
    - As a convenience, all of the inputs from the forward pass are additionally made
        available to you on the backward pass.

    !!! tip

        If you need gradients with respect to multiple arguments, then just pack them
        together as a tuple via the first argument `vjp_arg`. (See also
        [`equinox.filter_grad`][] for a similar trick.)
    """

    def __init__(self, fn):
        self.fn = fn
        self.fn_wrapped = None

    def defvjp(self, fn_fwd, fn_bwd):
        def fn_wrapped(
            nonarray_vjp_arg,
            nonarray_args_kwargs,
            diff_array_vjp_arg,
            nondiff_array_vjp_arg,
            array_args_kwargs,
        ):
            vjp_arg = combine(
                nonarray_vjp_arg, diff_array_vjp_arg, nondiff_array_vjp_arg
            )
            args, kwargs = combine(nonarray_args_kwargs, array_args_kwargs)
            out = self.fn(vjp_arg, *args, **kwargs)
            array_out, nonarray_out = partition(out, is_array)
            diff_array_out, nondiff_array_out = partition(array_out, is_inexact_array)
            return diff_array_out, nondiff_array_out, Static(nonarray_out)

        def fn_fwd_wrapped(
            nonarray_vjp_arg,
            nonarray_args_kwargs,
            diff_array_vjp_arg,
            nondiff_array_vjp_arg,
            array_args_kwargs,
        ):
            vjp_arg = combine(
                nonarray_vjp_arg, diff_array_vjp_arg, nondiff_array_vjp_arg
            )
            args, kwargs = combine(nonarray_args_kwargs, array_args_kwargs)
            out, residuals = fn_fwd(vjp_arg, *args, **kwargs)
            array_out, nonarray_out = partition(out, is_array)
            diff_array_out, nondiff_array_out = partition(array_out, is_inexact_array)
            out = diff_array_out, nondiff_array_out, Static(nonarray_out)
            return out, (
                residuals,
                diff_array_vjp_arg,
                nondiff_array_vjp_arg,
                array_args_kwargs,
            )

        def fn_bwd_wrapped(nonarray_vjp_arg, nonarray_args_kwargs, residuals, grad_out):
            (
                residuals,
                diff_array_vjp_arg,
                nondiff_array_vjp_arg,
                array_args_kwargs,
            ) = residuals
            vjp_arg = combine(
                nonarray_vjp_arg, diff_array_vjp_arg, nondiff_array_vjp_arg
            )
            args, kwargs = combine(nonarray_args_kwargs, array_args_kwargs)
            grad_diff_array_out, _, _ = grad_out
            out = fn_bwd(residuals, grad_diff_array_out, vjp_arg, *args, **kwargs)
            if jtu.tree_structure(out) != jtu.tree_structure(diff_array_vjp_arg):
                raise RuntimeError(
                    "custom_vjp gradients must have the same structure as "
                    "`equinox.filter(vjp_arg, equinox.is_inexact_array)`, where "
                    "`vjp_arg` is the first argument used in the forward pass."
                )
            # None is the gradient through nondiff_array_vjp_arg and array_args_kwargs
            return out, None, None

        fn_wrapped = jax.custom_vjp(fn_wrapped, nondiff_argnums=(0, 1))
        fn_wrapped.defvjp(fn_fwd_wrapped, fn_bwd_wrapped)
        self.fn_wrapped = fn_wrapped

    def __call__(__self, __vjp_arg, *args, **kwargs):
        # Try and avoid name collisions with the arguments of the wrapped function.
        # TODO: once we switch to Python 3.8, use (self, vjp_arg, /, *args, **kwargs).
        self = __self
        vjp_arg = __vjp_arg
        del __self, __vjp_arg
        if self.fn_wrapped is None:
            raise RuntimeError(f"defvjp not yet called for {self.fn.__name__}")
        array_vjp_arg, nonarray_vjp_arg = partition(vjp_arg, is_array)
        diff_array_vjp_arg, nondiff_array_vjp_arg = partition(
            array_vjp_arg, is_inexact_array
        )
        array_args_kwargs, nonarray_args_kwargs = partition((args, kwargs), is_array)
        out = self.fn_wrapped(
            nonarray_vjp_arg,
            nonarray_args_kwargs,
            diff_array_vjp_arg,
            nondiff_array_vjp_arg,
            array_args_kwargs,
        )
        diff_array_out, nondiff_array_out, nonarray_out = out
        return combine(diff_array_out, nondiff_array_out, nonarray_out.value)


if getattr(typing, "GENERATING_DOCUMENTATION", False):
    _filter_custom_jvp_doc = filter_custom_jvp.__doc__
    _filter_custom_vjp_doc = filter_custom_vjp.__doc__

    def defjvp(fn_jvp):
        pass

    def filter_custom_jvp(fn):
        return types.SimpleNamespace(defjvp=defjvp)

    def defvjp(fn_fwd, fn_bwd):
        pass

    def filter_custom_vjp(fn):
        return types.SimpleNamespace(defvjp=defvjp)

    filter_custom_jvp.__doc__ = _filter_custom_jvp_doc
    filter_custom_vjp.__doc__ = _filter_custom_vjp_doc



import functools as ft

import jax.tree_util as jtu

from .filters import combine, partition


def hashable_partition(pytree, filter_spec):
    dynamic, static = partition(pytree, filter_spec)
    static_leaves, static_treedef = jtu.tree_flatten(static)
    static_leaves = tuple(static_leaves)
    return dynamic, (static_leaves, static_treedef)


def hashable_combine(dynamic, static):
    static_leaves, static_treedef = static
    static = jtu.tree_unflatten(static_treedef, static_leaves)
    return combine(dynamic, static)


def _strip_wrapped_partial(fun):
    if hasattr(fun, "__wrapped__"):  # ft.wraps
        return _strip_wrapped_partial(fun.__wrapped__)
    if isinstance(fun, ft.partial):
        return _strip_wrapped_partial(fun.func)
    return fun


def get_fun_names(fun):
    fun = _strip_wrapped_partial(fun)
    try:
        return fun.__name__, fun.__qualname__
    except AttributeError:
        return type(fun).__name__, type(fun).__qualname__


def compile_cache(fun):
    @ft.lru_cache(maxsize=None)
    def _cache(leaves, treedef):
        args, kwargs = jtu.tree_unflatten(treedef, leaves)
        return fun(*args, **kwargs)

    @ft.wraps(fun)
    def _fun(*args, **kwargs):
        leaves, treedef = jtu.tree_flatten((args, kwargs))
        leaves = tuple(leaves)
        return _cache(leaves, treedef)

    return _fun



import functools as ft
from typing import Optional

import jax
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.interpreters.xla as xla
import jax.tree_util as jtu
from jaxtyping import PyTree

from ..filters import combine, is_array, partition
from .errors import error_if


_identity = lambda x, *, msg: x


_nondifferentiable_p = jax.core.Primitive("nondifferentiable")


def _nondifferentiable_batch(x, batch_axes, *, msg):
    (x,) = x
    (batch_axes,) = batch_axes
    return nondifferentiable(x, msg=msg), batch_axes


def _nondifferentiable_jvp(primals, tangents, *, msg):
    raise RuntimeError(msg)


_nondifferentiable_p.def_impl(_identity)
_nondifferentiable_p.def_abstract_eval(_identity)
batching.primitive_batchers[_nondifferentiable_p] = _nondifferentiable_batch
if hasattr(xla, "lower_fun"):
    xla.register_translation(
        _nondifferentiable_p,
        xla.lower_fun(_identity, multiple_results=False, new_style=True),
    )
mlir.register_lowering(
    _nondifferentiable_p,
    mlir.lower_fun(_identity, multiple_results=False),
)
ad.primitive_jvps[_nondifferentiable_p] = _nondifferentiable_jvp


def nondifferentiable(
    x: PyTree, *, name: Optional[str] = None, msg: Optional[str] = None
) -> PyTree:
    """
    Consumes a PyTree with arbitrary leaves and returns an identical PyTree.
    If any of the JAX arrays in this PyTree are ever differentiated (in
    forward or reverse mode) then an error will be thrown.
    """
    dynamic, static = partition(x, is_array)
    flat, treedef = jtu.tree_flatten(dynamic)
    if msg is None:
        if name is None:
            name = "This operation"
        msg = f"Unexpected tangent. {name} cannot be autodifferentiated."
    bind = ft.partial(_nondifferentiable_p.bind, msg=msg)
    flat = map(bind, flat)
    return combine(jtu.tree_unflatten(treedef, flat), static)


_nondifferentiable_backward_p = jax.core.Primitive("nondifferentiable_backward")


def _nondifferentiable_backward_batch(x, batch_axes, *, msg):
    (x,) = x
    (batch_axes,) = batch_axes
    return nondifferentiable_backward(x, msg=msg), batch_axes


def _nondifferentiable_backward_jvp(primals, tangents, *, msg):
    (primals,) = primals
    (tangents,) = tangents
    return nondifferentiable_backward(primals, msg=msg), nondifferentiable_backward(
        tangents, msg=msg
    )


def _nondifferentiable_backward_transpose(cts_in, _, *, msg):
    if isinstance(cts_in, ad.Zero):
        return ad.Zero  # the class, not an instance
    else:
        # Unfortunately there are legitimate cases where we get all-zero non-symbolic
        # cotangents, so we have to use a runtime error here instead of just erroring
        # at trace time.
        # This happens when doing something like:
        #
        # x = nondifferentiable_backward(x)
        # x, y = lax.cond(pred, lambda: (x, y), lambda: (x+1, y+1))
        # return y
        return [error_if(cts_in, (cts_in != 0).any(), msg)]


_nondifferentiable_backward_p.def_impl(_identity)
_nondifferentiable_backward_p.def_abstract_eval(_identity)
batching.primitive_batchers[
    _nondifferentiable_backward_p
] = _nondifferentiable_backward_batch
if hasattr(xla, "lower_fun"):
    xla.register_translation(
        _nondifferentiable_backward_p,
        xla.lower_fun(_identity, multiple_results=False, new_style=True),
    )
mlir.register_lowering(
    _nondifferentiable_backward_p,
    mlir.lower_fun(_identity, multiple_results=False),
)
ad.primitive_jvps[_nondifferentiable_backward_p] = _nondifferentiable_backward_jvp
ad.primitive_transposes[
    _nondifferentiable_backward_p
] = _nondifferentiable_backward_transpose


def nondifferentiable_backward(
    x: PyTree, name: Optional[str] = None, msg: Optional[str] = None
) -> PyTree:
    """
    Consumes a PyTree with arbitrary leaves and returns an identical PyTree.
    If any of the JAX arrays in this PyTree are ever differentiated in
    reverse mode then an error will be thrown.
    """
    dynamic, static = partition(x, is_array)
    flat, treedef = jtu.tree_flatten(dynamic)
    if msg is None:
        if name is None:
            name = "This operation"
        msg = f"Unexpected cotangent. {name} cannot be reverse-mode autodifferentiated."
    bind = ft.partial(_nondifferentiable_backward_p.bind, msg=msg)
    flat = map(bind, flat)
    return combine(jtu.tree_unflatten(treedef, flat), static)
